/*
 * IP_VS_CB:        Round-Robin Scheduling module
 *
 * Version:     $Id: ip_vs_rr.c,v 1.5 1999/10/06 10:04:46 wensong Exp $
 *
 * Authors:     Wensong Zhang <wensong@iinchina.net>
 *              Peter Kese <peter.kese@ijs.si>
 *
 *              This program is free software; you can redistribute it and/or
 *              modify it under the terms of the GNU General Public License
 *              as published by the Free Software Foundation; either version
 *              2 of the License, or (at your option) any later version.
 *
 * Fixes/Changes:
 *     Wensong Zhang            :     changed the ip_vs_rr_schedule to return dest
 *     Julian Anastasov         :     fixed the NULL pointer access bug in debugging
 *     Wensong Zhang            :     changed some comestics things for debugging
 *     Wensong Zhang            :     changed for the d-linked destination list
 *     Wensong Zhang            :     added the ip_vs_rr_update_svc
 *     Wensong Zhang            :     added any dest with weight=0 is quiesced 
 *
 */

#include <linux/config.h>
#include <linux/module.h>
#ifdef CONFIG_KMOD
#include <linux/kmod.h>
#endif
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <net/ip_masq.h>
#ifdef CONFIG_IP_MASQUERADE_MOD
#include <net/ip_masq_mod.h>
#endif
#include <linux/sysctl.h>
#include <linux/ip_fw.h>
#include <net/ip_vs.h>
#include <net/ip_mycs.h>
#include <net/ip_vs_cb_rule.h>

static struct server server_list[] = {
  {"ace", 0x80c6c0c6},
  {"frodo", 0x80c6c0c1},
  {"oblib", 0x80c6c0c3},
  {"gandalf", 0x80c6c0c2},
  {"default", 0x00000000},
  {"vinci", 0x80c6c0b7},
  {"reject", 0xffffffff} 
};

extern struct ip_vs_cb_rule_field rule_fields[];
extern int num_rule_fields;

int num_servers = sizeof(server_list) / sizeof(struct server);	

__u32 check_sticky(__u32 addr)
{
  int i = 0;

  for(i = 0; i < sticky_num; i++)
  {
    if(addr == ip_sticky_list[i].source_ip)
    {
      IP_RULE_MSG("sticky found on %x\n",ip_sticky_list[i].dest_ip);
      return ip_sticky_list[i].dest_ip;
    }
  }
  return 0;
}

__u32 route_to(char *server_name, int is_sticky)
{
  int i = 0;
  __u32 dest_addr = 0;

  for(i = 0; i < num_servers; i++)
  {
    if(strcmp(server_name, server_list[i].name) == 0)
    {
      if(is_sticky)
      {
	/*printk("###### SERVER_ROUTE_TO=  STICKY  \n");*/
	ip_sticky_list[sticky_num].dest_ip = server_list[i].addr;
	ip_sticky_list[sticky_num].source_ip = source_ip;
	sticky_num++;
	/*printk("##### in route_to sticky_num=%d\n",sticky_num);*/
      }
      dest_addr = server_list[i].addr;
      break;
    }
  }

  return dest_addr;
}

__u32 module_rule_matching(struct sk_buff *skb)
{
  struct iphdr *iph=skb->nh.iph;
  int data_len = 0;
  unsigned char *data_p = NULL;
  int i = 0;
  __u32 ret_addr = 0;
  __u32 saddr = 0, daddr = 0;
  __u16 sport = 0, dport = 0;
  __u8 protocol = 0;
  union ip_masq_tphdr h;
  void (*field_parser) (const unsigned char *data,  int data_len);

  field_parser = NULL;
  
  MOD_INC_USE_COUNT;
  
  h.raw = (char*) iph + iph->ihl * 4;
  
  saddr = ntohl(iph->saddr);
  daddr = ntohl(iph->daddr);
  sport = ntohs(h.th->source);
  dport = ntohs(h.th->dest);
  protocol = iph->protocol;
  source_ip = saddr;
  
  IP_RULE_MSG("saddr=%x daddr=%x, sport=%d dport =%d proto=%d\n", saddr, daddr, sport, dport, protocol);

  if((ret_addr = check_sticky(saddr)) > 0)
    goto return_directly;

  data_p = skb->data + iph->ihl*4 + skb->h.th->doff*4;
  data_len = skb->len - (iph->ihl*4 + skb->h.th->doff*4);

  if(data_len <= 0)
  {
    IP_RULE_MSG("No data found data len = %d", data_len);
    goto return_directly;
  }
  
  if(data_len >= 4)
    IP_RULE_MSG("1st4Bytes = %c%c%c%c\n", data_p[0], data_p[2], data_p[2], data_p[3]);

  field_parser = get_content_type(data_p, data_len);
  
  if(field_parser)
  {
    /*
     * This function populates all the fields specified in the rules
     * with data obtained from the incoming request
     */

    field_parser(data_p, data_len);

    ret_addr = rule_configure(saddr, daddr, sport, dport, protocol);
  }

return_directly:
  MOD_DEC_USE_COUNT;

  IP_RULE_MSG("Returned address=%x\n", ret_addr);

  return ret_addr;
}

void register_rule_matching()
{
  rule_matching = module_rule_matching;
}

void unregister_rule_matching()
{
  rule_matching = ip_cs_norule;
}

int init_module(void)
{
  register_rule_matching();
  
  IP_RULE_MSG("RULE MATCHING MODULE INSERTED num_servers=%d num_rule_fields=%d\n", num_servers, num_rule_fields);

  return 0;
}

void cleanup_module(void)
{
  unregister_rule_matching();
  
  IP_RULE_MSG("RULE MATCHING MODULE REMOVED\n");
}
